import { db } from "$lib/server/db";
import * as table from "$lib/server/db/schema";
import { connectedDevices, deviceSensorData, getCurrentDevices } from "$lib/server/mqtt-devices.js";
import { eq } from "drizzle-orm";
import mqtt from "mqtt";
import { writable } from "svelte/store";

const latestMessage = writable("No message yet");
const devices = writable([]);

let client = null;

function updateDevice(deviceId, sensorData = null) {
  const now = new Date();
  const deviceName = `ESP8266 #${deviceId.slice(-6)}`;

  connectedDevices.set(deviceId, {
    id: deviceId,
    name: deviceName,
    type: "esp8266",
    status: "online",
    lastSeen: now,
  });

  if (sensorData) {
    deviceSensorData.set(deviceId, {
      ...sensorData,
      timestamp: now,
    });
  }

  updateDevicesStore();
}

function parseSensorData(payload) {
  try {
    const data = JSON.parse(payload);
    return {
      temperature: parseFloat(data.temperature) || 0,
      humidity: parseFloat(data.humidity?.replace("%", "")) || 0,
      pressure: parseFloat(data.pressure) || 0,
      altitude: parseFloat(data.altitude) || 0,
    };
  } catch (error) {
    console.error("Error parsing sensor data:", error);
    return null;
  }
}

const sseControllers = new Set();

function updateDevicesStore() {
  const deviceList = Array.from(connectedDevices.values()).map((device) => {
    const sensorData = deviceSensorData.get(device.id);
    return {
      ...device,
      sensorData: sensorData || null,
    };
  });
  devices.set(deviceList);
}

function cleanupDevices() {
  const fiveMinutesAgo = new Date(Date.now() - 5 * 60 * 1000);
  let updated = false;

  for (const [deviceId, device] of connectedDevices.entries()) {
    if (device.lastSeen < fiveMinutesAgo && device.status === "online") {
      connectedDevices.set(deviceId, { ...device, status: "offline" });
      updated = true;
    }
  }

  if (updated) {
    updateDevicesStore();
  }
}

setInterval(cleanupDevices, 60000);

function connectMqtt() {
  if (client && client.connected) {
    return;
  }

  const BROKER_URL = "mqtt://kada49.it:1883";
  const TOPIC = "esp8266/+/data";

  client = mqtt.connect(BROKER_URL);

  client.on("connect", () => {
    console.log("Connected to MQTT broker!");
    client.subscribe(TOPIC, (err) => {
      if (!err) {
        console.log(`Subscribed to topic: ${TOPIC}`);
      } else {
        console.error(`Subscription error: ${err}`);
      }
    });
  });

  client.on("message", async (topic, message) => {
    const payload = message.toString();
    console.log(`Received message from topic "${topic}": ${payload}`);
    latestMessage.set(payload);

    const topicParts = topic.split("/");
    if (topicParts.length >= 3 && topicParts[0] === "esp8266" && topicParts[2] === "data") {
      const deviceId = topicParts[1];
      console.log(`Processing data for device: ${deviceId}`);

      const sensorData = parseSensorData(payload);
      if (sensorData) {
        console.log(`Parsed sensor data:`, sensorData);
        updateDevice(deviceId, sensorData);
        const devices = await db.select().from(table.sensors).where(eq(table.sensors.id, deviceId));
        if (devices.length == 1)
          await db.insert(table.sensorData).values({
            sensor: deviceId,
            temperature: sensorData.temperature,
            humidity: sensorData.humidity,
            altitude: sensorData.altitude,
            pressure: sensorData.pressure,
          });
      } else {
        updateDevice(deviceId);
      }
    }
  });

  client.on("error", (err) => {
    console.error(`MQTT error: ${err}`);
    if (client) {
      client.end();
    }
    client = null;
  });

  client.on("close", () => {
    console.log("MQTT connection closed.");
    client = null;
  });
}

connectMqtt();

/** @type {import("./$types").RequestHandler} */
export async function GET({ request }) {
  const headers = {
    "Content-Type": "text/event-stream",
    "Cache-Control": "no-cache",
    Connection: "keep-alive",
  };

  return new Response(
    new ReadableStream({
      start(controller) {
        console.log("Client connected to MQTT SSE stream");

        let isConnected = true;
        let interval;

        sseControllers.add(controller);

        const cleanup = () => {
          console.log("Cleaning up MQTT SSE connection");
          isConnected = false;
          if (interval) {
            clearInterval(interval);
            interval = null;
          }
          sseControllers.delete(controller);
          if (unsubscribe) {
            unsubscribe();
          }
        };

        function sendDevices() {
          if (!isConnected) {
            cleanup();
            return;
          }

          try {
            const devices = getCurrentDevices();
            const message = {
              devices: devices,
              timestamp: new Date().toISOString(),
            };

            if (isConnected) {
              controller.enqueue(`data: ${JSON.stringify(message)}\n\n`);
            }
          } catch (error) {
            if (
              error.code === "ERR_INVALID_STATE" ||
              error.message.includes("Controller is already closed")
            ) {
              console.log("MQTT SSE controller closed, stopping interval");
            } else {
              console.error("Error sending MQTT device data:", error);
            }
            cleanup();
          }
        }

        sendDevices();

        interval = setInterval(sendDevices, 2000);

        const unsubscribe = latestMessage.subscribe((message) => {
          if (message !== "No message yet" && isConnected) {
            try {
              const data = `data: ${JSON.stringify({
                message: message,
                devices: getCurrentDevices(),
                timestamp: new Date().toISOString(),
              })}\n\n`;
              if (isConnected) {
                controller.enqueue(data);
              }
            } catch (error) {
              if (
                error.code === "ERR_INVALID_STATE" ||
                error.message.includes("Controller is already closed")
              ) {
                console.log("MQTT SSE controller closed during message send");
                cleanup();
              }
            }
          }
        });

        request.signal.addEventListener("abort", () => {
          console.log("Client disconnected from MQTT SSE stream");
          cleanup();
        });
      },
      cancel() {
        console.log("MQTT SSE stream cancelled");
      },
    }),
    { headers },
  );
}
